@@ -7,6 +7,8 @@ from rich.logging import RichHandler
 
 from . import msm_we
 from sklearn.cluster import KMeans
+import pickle
+import base64
 
 log = logging.getLogger(__name__)
 log.setLevel(logging.INFO)
@@ -61,7 +63,7 @@ def solve_discrepancy(tmatrix, pi, B):
     return discrepancy, variance
 
 
-def get_uniform_mfpt_bins(variance, steady_state, n_desired_we_bins):
+def get_uniform_mfpt_bins(variance, discrepancy, steady_state, n_desired_we_bins):
     """
     Implements the MFPT-binning strategy described in [1], where bins are groups of microstates that are uniformly
     spaced in the integral of pi * v
@@ -84,6 +86,8 @@ def get_uniform_mfpt_bins(variance, steady_state, n_desired_we_bins):
 
     # The last two elements of this are the basis and target states respectively
     pi_v = steady_state * variance
+    pi_v_sort = np.argsort(discrepancy).squeeze()
+    cumsum = np.cumsum(pi_v[pi_v_sort])
 
     spacing = sum(pi_v) / n_desired_we_bins
 
@@ -92,10 +96,10 @@ def get_uniform_mfpt_bins(variance, steady_state, n_desired_we_bins):
         lower, upper = spacing * i, spacing * (i+1)
         log.debug(f"Checking for states with pi_v between {lower}, {upper}")
 
-        states_in_bin = np.argwhere(
-            (lower < np.cumsum(pi_v)) &
-            (np.cumsum(pi_v) <= upper)
+        indices_in_bin = np.argwhere(
+            (lower < cumsum) & (cumsum <= upper)
         )
+        states_in_bin = pi_v_sort[indices_in_bin]
 
         log.info(f"Found that bin {i} contains microstates {states_in_bin}")
         bin_states[states_in_bin] = i
@@ -145,7 +149,31 @@ def get_clustered_mfpt_bins(variance, steady_state, n_desired_we_bins):
 
 class OptimizedBinMapper(westpa.core.binning.FuncBinMapper):
 
-    def __init__(self,
+    def __init__(self, *args, **kwargs):
+
+        if len(args) + len(kwargs) == 1:
+            log.info("Single argument provided to binmapper initializer, attempting to unpickle")
+            self.unpickle(kwargs['bytestring'])
+
+        elif len(args) + len(kwargs) > 1:
+            log.info("Multiple arguments provided to binmapper initializer, creating new object")
+            self.create_new(*args, **kwargs)
+
+    def unpickle(self, bytestring):
+
+        # TODO: This may need to be provided as a b64-encoded bytestring, rather than plain bytes,
+        #  in which case we need to quickly encode it back to bytes. This can be done with
+        #   pickle.loads(base64.b64decode(collection.metadata['bin_mapper']))
+
+        loaded = pickle.loads(base64.b64decode(bytestring))
+
+        assert type(loaded) is type(self), "Invalid pickled object provided"
+
+        # TODO: Probably just update the attributes I explicitly care about, this seems a little insecure
+        for k, v in vars(loaded).items():
+            setattr(self, k, v)
+
+    def create_new(self,
                  nbins: int,
                  n_original_pcoord_dims: int,
                  target_pcoord_bounds,
@@ -161,15 +189,9 @@ class OptimizedBinMapper(westpa.core.binning.FuncBinMapper):
 
         Parameters
         ----------
-        nbins, int: Number of WE bins
-        n_original_pcoord_dims, int: Number of dimensions in the original user-supplied progress coordinate
-        cluster_centers, array-like: Array of microstate cluster centers
-        microstate_mapper, dict: Mapping of microstates to WE bins
-
-        TODO
-        ----
-        - Should we just have a separate dedicated target bin?
-        -
+        nbins: int, Number of WE bins
+        n_original_pcoord_dims: int, Number of dimensions in the original user-supplied progress coordinate
+        microstate_mapper: dict, Mapping of microstates to WE bins
         """
 
         super().__init__(func=self.mapper, nbins=nbins, args=args, kwargs=kwargs)
